/*
 *  XMMS Input Plugin - CDDA Paranoia
 *
 *  Copyright (C) 2002  Iwasa Kazmi 
 *
 *  Copyright (C) 1998-2000  Peter Alm, Mikael Alm, Olle Hallnas,
 *                           Thomas Nilsson and 4Front Technologies
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"

#include <sys/types.h>
#if HAVE_STRING_H
#include <string.h>
#endif
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#include <stdio.h>
#include <pthread.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <cdda_interface.h>
#include <cdda_paranoia.h>
#include <glib.h>


#include "cdparanoia.h"


static void cdparanoia_init(void);
static int cdparanoia_is_our_file(char *);
static GList *cdparanoia_scan_dir(char *);
static void cdparanoia_play_file(char *);
static void cdparanoia_stop(void);
static void cdparanoia_pause(short);
static void cdparanoia_seek(int);
static void cdparanoia_get_song_info(char *, char **, int *);
static int cdparanoia_get_time(void);
static void cdparanoia_cleanup(void);

InputPlugin cdparanoia_ip =
{
    NULL,
    NULL,
    NULL, /* Description */
    cdparanoia_init,
    NULL, /* about */
    cdparanoia_configure,
    cdparanoia_is_our_file,
    cdparanoia_scan_dir,
    cdparanoia_play_file,
    cdparanoia_stop,
    cdparanoia_pause,
    cdparanoia_seek,
    NULL,
    cdparanoia_get_time,
    NULL,
    NULL,
    cdparanoia_cleanup,
    NULL,
    NULL,
    NULL,
    NULL,
    cdparanoia_get_song_info,
    NULL, /* file_info_box */
    NULL
};


/*XXX this value must be variable */
#define DEFAULT_PARANOIA_MODE	(PARANOIA_MODE_FULL^PARANOIA_MODE_NEVERSKIP)


typedef struct
{
    cdrom_drive *d;
    long sec_top;
    long sec_last;
} PlayTrackStruc;

typedef struct
{
    long sec_top;
    long sec_last;
} TrackInfoCache;


static pthread_t play_thread;
static int playing = 0;
static volatile int stop_f = 0;
static volatile int seek_to = -1;
static TrackInfoCache *trackinfo_cache = NULL;
static int trackinfo_cache_count = 0;


InputPlugin *get_iplugin_info(void)
{
    cdparanoia_ip.description = g_strdup_printf(_("CDDA Paranoia %s"), VERSION);
    return &cdparanoia_ip;
}



static cdrom_drive *open_cdparanoia(void)
{
    cdrom_drive *d = NULL;
    int r;

    d = cdda_identify(para_cfg.cdrom_device, CDDA_MESSAGE_FORGETIT, NULL);
    if (d) {
        cdda_verbose_set(d, CDDA_MESSAGE_FORGETIT, CDDA_MESSAGE_FORGETIT);
        r = cdda_open(d);
        if (r) {
            cdda_close(d);
            d = NULL;
        }
    }

    return d;
}

static void setup_cdrom_drive(cdrom_drive *d)
{
    if (d) {
        if (para_cfg.force_cdrom_speed && para_cfg.cdrom_speed > 0)
            cdda_speed_set(d, para_cfg.cdrom_speed);
    }
}

static void close_cdparanoia(cdrom_drive *d)
{
    if (d) {
        cdda_close(d);
        d = NULL;
    }
}

static void cleanup_trackinfo_cache(void)
{
    if (trackinfo_cache) {
        g_free(trackinfo_cache);
        trackinfo_cache = NULL;
        trackinfo_cache_count = 0;
    }
}
 
static void allocate_trackinfo_cache(int t)
{
    cleanup_trackinfo_cache();
    trackinfo_cache = g_new0(TrackInfoCache, t);
    if (trackinfo_cache)
        trackinfo_cache_count = t;
}


static void cdparanoia_init(void)
{
    load_cfg();
}


static GList *cdparanoia_scan_dir(char *dir)
{
    cdrom_drive *d = NULL;
    GList *list = NULL;
    gint i;

    cleanup_trackinfo_cache();

    d = open_cdparanoia();
    if (d) {
        if (strncmp(para_cfg.cdrom_dir, dir, strlen(para_cfg.cdrom_dir)) == 0) {
            if (d->tracks > 0) {
                allocate_trackinfo_cache(d->tracks);
            }
            for(i = 1; i <= d->tracks; i++) {
                if (cdda_track_audiop(d, i)) {
                    if (trackinfo_cache) {
                        trackinfo_cache[i-1].sec_top = cdda_track_firstsector(d, i);
                        trackinfo_cache[i-1].sec_last = cdda_track_lastsector(d, i);
                    }
                    list = g_list_append(list, g_strdup_printf("Track %02d.cdp", i));
                }
            }
        }
    }
    close_cdparanoia(d);

    return list;
}

static int cdparanoia_is_our_file(char *filename)
{
    gchar *ext;

    ext = strrchr(filename, '.');
    if (ext && !strcasecmp(ext, ".cdp"))
        return TRUE;

    return FALSE;
}

static int get_track_number(char *filename)
{
    char *tmp;
    int i;

    tmp = strrchr(filename, '/');
    if (tmp)
        tmp++;
    else
        tmp = filename;
        
    if (!sscanf(tmp, "Track %d.cdp", &i))
        return -1;

    return i;
}

static void cdparanoia_get_song_info(char *filename, char **title, int *len)
{
    long sec_top;
    long sec_last;
    int i;

    *title = NULL;
    *len = -1;

    i = get_track_number(filename);
    if (i < 1 || i > trackinfo_cache_count)
        return;

    sec_top = trackinfo_cache[i-1].sec_top;
    sec_last = trackinfo_cache[i-1].sec_last;
    *len = (sec_last - sec_top + 1) * 1000 / 75;
    *title = g_strdup_printf(_("CD Audio Track %02d"), i);
}

static void paranoia_callback(long inpos, int function)
{
    /**/
}

/* thread function */
static void *play_loop(void *arg)
{
    PlayTrackStruc ptrack = *(PlayTrackStruc*)arg; /*copy*/
    cdrom_paranoia *p = NULL;
    long csec;
    char *err;
    char *mes;
    int16_t *readbuf = NULL;

    p = paranoia_init(ptrack.d);
    paranoia_modeset(p, DEFAULT_PARANOIA_MODE);
    paranoia_seek(p, ptrack.sec_top, SEEK_SET);
    csec = ptrack.sec_top;

    /* cdparanoia do this */
    seteuid(getuid());
    setegid(getgid());

    while(!stop_f) {
        if (seek_to >= 0) {
            long sec;
            cdparanoia_ip.output->flush(seek_to * 1000);
            sec = ptrack.sec_top + seek_to * 75;
            if (sec <= ptrack.sec_last) {
                paranoia_seek(p, sec, SEEK_SET);
            }
            csec = sec;
            readbuf = NULL;
            seek_to = -1;
        }

        if (!readbuf) {
            int abort_f = 0;
            if (csec <= ptrack.sec_last) {
                readbuf = paranoia_read_limited(p, paranoia_callback, para_cfg.paranoia_retries);
                err = cdda_errors(ptrack.d);
                mes = cdda_messages(ptrack.d);
                if (err || mes) {
                    free(err);
                    free(mes);
                }
                if (!readbuf && para_cfg.abort_on_skip) {
                    abort_f = 1;
                }
                csec++;
            } else {
                abort_f = 1;
            }

            if (abort_f) {
                while(cdparanoia_ip.output->buffer_playing()) {
                    cdparanoia_ip.output->buffer_free();
                    xmms_usleep(20000);
                }
                break;
            }
        }

        if (readbuf && cdparanoia_ip.output->buffer_free() >= CD_FRAMESIZE_RAW) {
            cdparanoia_ip.add_vis_pcm(
                    cdparanoia_ip.output->written_time(),
                    FMT_S16_LE,
                    2, CD_FRAMESIZE_RAW, readbuf);

            cdparanoia_ip.output->write_audio(readbuf, CD_FRAMESIZE_RAW);
            readbuf = NULL;
        } else {
            xmms_usleep(20000);
        }
    }

    paranoia_free(p);
    close_cdparanoia(ptrack.d);

    playing = 0;
    cdparanoia_ip.output->close_audio();
    pthread_exit(NULL);
}

static void cdparanoia_play_file(char *filename)
{
    static PlayTrackStruc ptrack;
    int i;

    if (!playing) {
        ptrack.d = open_cdparanoia();
        if (!ptrack.d)
            return;

        i = get_track_number(filename);

        if (i < 1 || i > ptrack.d->tracks) {
            close_cdparanoia(ptrack.d);
            return;
        }

        setup_cdrom_drive(ptrack.d);

        if (cdparanoia_ip.output->open_audio(FMT_S16_LE, 44100, 2) != 0) {
            gchar *name;
            ptrack.sec_top = cdda_track_firstsector(ptrack.d, i);
            ptrack.sec_last = cdda_track_lastsector(ptrack.d, i);
            name = g_strdup_printf("Track %02d", i);
            cdparanoia_ip.set_info(name, (ptrack.sec_last - ptrack.sec_top + 1) * 1000 / 75, 16 * 44100 * 2, 44100, 2);
            g_free(name);
            playing = 1;
            stop_f = 0;
            seek_to = -1;
            if (pthread_create(&play_thread, NULL, play_loop, &ptrack) != 0)
                playing = 0;
        }

        if (!playing) {
            close_cdparanoia(ptrack.d);
        }
    }
}

static void cdparanoia_stop(void)
{
    if (playing) {
        stop_f = 1;
        pthread_join(play_thread, NULL);
    }
}

static void cdparanoia_pause(short p)
{
    cdparanoia_ip.output->pause(p);
}

static void cdparanoia_seek(int time)
{
    if (playing) {
        seek_to = time;
        while (seek_to >= 0)
            xmms_usleep(10000);
    }
}

static int cdparanoia_get_time(void)
{
    if (!playing)
        return -1;

    return cdparanoia_ip.output->output_time();
}

static void cdparanoia_cleanup(void)
{
    cleanup_trackinfo_cache();
}



